"""
Lightning callback to inspect model generations during training.
"""

import hashlib
from pathlib import Path

import lightning as L
import torch
from lightning.pytorch.callbacks import BasePredictionWriter

from gpt2_prot.data_module import AATokenizer, NTTokenizer
from gpt2_prot.gpt2 import PredictConfig


class PreviewCallback(L.Callback):
    """
    Callback to print a preview of the models generation.

    Params:
        mode (str): Run in protein (aa) or nucleotide mode.
        prompt (str): The prompt to generate from.
        length (int): Sequence length to generate.
    Returns:
        None
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, mode: str, prompt: str, length: int = 50) -> None:
        super().__init__()
        assert mode in ["aa", "nt"]
        self.mode = mode
        self.prompt = prompt
        self.length = length

    def on_train_epoch_end(self, trainer: L.Trainer, pl_module: L.LightningModule) -> None:
        """
        Logic to run the model generation, additionally logs to tensorboard.

        Params:
            trainer (L.Trainer): The trainer object.
            pl_module (L.LightningModule): The LightningModule object.
        Returns:
            None
        """
        pl_module.eval()

        if self.mode == "aa":
            tok = AATokenizer()
        else:
            tok = NTTokenizer()

        prompt_enc = torch.LongTensor(tok(self.prompt)).unsqueeze(0).to(pl_module.device)

        generate_arguments = [
            {"t": 1.0, "sample": False, "top_k": None},
            {"t": 1.0, "sample": True, "top_k": None},
            {"t": 1.2, "sample": True, "top_k": 5},
        ]

        tag = "Model generation previews:"
        message = ""

        for args in generate_arguments:
            response_enc = pl_module.generate(prompt_enc, self.length, **args)
            seq = tok.decode(response_enc.flatten().tolist())
            message += f"({args}): {seq}\n"

        tensorboard = pl_module.logger.experiment  # type: ignore
        tensorboard.add_text(tag, message, trainer.global_step)

        print(tag, "\n", message)


class FastaInferenceWriter(BasePredictionWriter):
    """
    Callback to configure the model and write generated sequences to a fasta file.

    Params:
        output_file (Path): The path to the output file.
        mode (str): Run in protein (aa) or nucleotide mode.
        max_tokens (int): Maximum number of tokens to generate per sequence.
        t (float): Temperature parameter to use during inference.
        sample (bool): Whether to sample from model logits during inference.
        top_k (int | None): Top k parameter to use during inference.
    Returns:
        None
    """

    # pylint: disable=abstract-method, too-many-arguments
    def __init__(
        self,
        output_file: Path,
        mode: str,
        max_tokens: int,
        t: float,
        sample: bool,
        top_k: int | None,
    ):
        super().__init__(write_interval="epoch")
        self.output_file = output_file
        assert mode in ["aa", "nt"]
        self.mode = mode

        self.predict_config = PredictConfig(max_tokens=max_tokens, t=t, sample=sample, top_k=top_k)

        self.output_file.parent.mkdir(parents=True, exist_ok=True)
        self.output_file.touch(exist_ok=True)

    def on_predict_start(self, trainer: L.Trainer, pl_module: L.LightningModule):
        """
        Configure the model with inference parameters before predicting.

        Params:
            trainer (L.Trainer): The trainer object (UNUSED).
            pl_module (L.LightningModule): The LightningModule object.
        Returns:
            None
        """
        _ = trainer
        pl_module.predict_config = self.predict_config

    @staticmethod
    def _hash_sequence(sequence: str) -> str:
        return hashlib.sha256(sequence.encode("utf-8")).hexdigest()[:8]

    def write_on_epoch_end(
        self,
        trainer: L.Trainer,
        pl_module: L.LightningModule,
        predictions: list[torch.Tensor],
        batch_indices: list[int],
    ):
        """
        Decode the tokenised model generations and write to a fasta file.

        Params:
            trainer (L.Trainer): The trainer object (UNUSED).
            pl_module (L.LightningModule): The LightningModule object (UNUSED).
            predictions (list[torch.Tensor]): The model predictions.
            batch_indices (list[int]): The batch indices (UNUSED).
        Returns:
            None
        """
        _ = trainer, pl_module, batch_indices

        if self.mode == "aa":
            tok = AATokenizer()
        else:
            tok = NTTokenizer()

        out_file = self.output_file.open("a")

        for pred in predictions:
            seq = tok.decode(pred.flatten())
            print(f">{self._hash_sequence(seq)}\n{seq}", file=out_file)

        out_file.close()


if __name__ == "__main__":
    pass



"""
Main entry point for gpt2-prot.
"""

from lightning.pytorch.cli import LightningCLI

from gpt2_prot.callbacks import PreviewCallback  # pylint: disable=unused-import
from gpt2_prot.data_module import BioDataModule
from gpt2_prot.gpt2 import GPT2


def main() -> None:
    """
    Lightning CLI entry point.
    """
    LightningCLI(GPT2, BioDataModule)


if __name__ == "__main__":
    main()



"""
Unit tests for the package using pytest.

REF: <https://docs.pytest.org/en/7.4.x/>
"""


def test_imports():
    """Test that the package and a submodule can be imported."""

    import gpt2_prot  # pylint: disable=import-outside-toplevel

    assert gpt2_prot

    import gpt2_prot.scripts  # pylint: disable=import-outside-toplevel

    assert gpt2_prot.scripts



"""
General purpose GPT2 implementation in pytorch lightning.
"""

from math import sqrt
from typing import NamedTuple

import lightning as L
import torch
from torch import nn
from torch.nn import functional as F
from torcheval.metrics.functional.text import perplexity


class GPT2Config(NamedTuple):
    """Hyperparameters for GPT2."""

    vocab_size: int
    window_size: int

    n_layers: int
    n_heads: int
    embed_d: int

    emb_dropout: float
    attn_dropout: float
    res_dropout: float

    adam_lr: float
    adam_weight_decay: float
    adam_betas: tuple[float, float]


class PredictConfig(NamedTuple):
    """Parameters for model generations."""

    max_tokens: int = 100
    sample: bool = False
    t: float = 1.0
    top_k: int | None = None


class TransformerBlock(nn.Module):
    """
    Decoder only GPT2 transformer block.

    Params:
        embed_d (int): Internal embedding dimension.
        n_heads (int): Number of attention heads.
        attn_dropout (float): Dropout rate for attention weights.
        res_dropout (float): Dropout rate for residual connections.
    Returns:
        None
    """

    # pylint: disable=too-few-public-methods

    def __init__(
        self,
        embed_d: int,
        n_heads: int,
        attn_dropout: float,
        res_dropout: float,
    ) -> None:
        super().__init__()

        self.ln1 = nn.LayerNorm(embed_d)
        self.qkv = nn.Linear(embed_d, 3 * embed_d)
        self.attn = nn.MultiheadAttention(embed_d, n_heads, attn_dropout, batch_first=True)
        self.ln2 = nn.LayerNorm(embed_d)
        self.mlp = nn.Sequential(
            nn.Linear(embed_d, 4 * embed_d),
            nn.GELU(),
            nn.Linear(4 * embed_d, embed_d),
            nn.Dropout(res_dropout),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass through the transformer block."""
        q, k, v = self.qkv(self.ln1(x)).split(self.ln1.normalized_shape[0], dim=2)
        mask = self._lookahead_mask(x.size(1)).to(x.device) if self.training else None
        x_, _ = self.attn(q, k, v, attn_mask=mask, need_weights=False)
        x = x + x_
        x = x + self.mlp(self.ln2(x))
        return x

    @staticmethod
    def _lookahead_mask(d: int) -> torch.Tensor:
        """Generate a lookahead attention mask for a given dimension."""
        mask = torch.triu(torch.ones(d, d), diagonal=1)
        mask[mask.bool()] = -float("inf")
        return mask


class GPT2(L.LightningModule):
    """
    GPT2 language model.

    Params:
        config (GPT2Config): Hyperparameters for GPT2.
    Returns:
        None
    """

    def __init__(self, config: GPT2Config) -> None:
        super().__init__()
        self.config = config
        self.predict_config: PredictConfig = PredictConfig()

        transformer_blocks = [
            TransformerBlock(
                config.embed_d,
                config.n_heads,
                config.attn_dropout,
                config.res_dropout,
            )
            for _ in range(config.n_layers)
        ]

        self.transformer = nn.ModuleDict(
            {
                "tok_emb": nn.Embedding(config.vocab_size, config.embed_d),
                "pos_emb": nn.Embedding(config.window_size, config.embed_d),
                "dropout": nn.Dropout(config.emb_dropout),
                "blocks": nn.ModuleList(transformer_blocks),
                "ln": nn.LayerNorm(config.embed_d),
            }
        )
        self.lm_head = nn.Linear(config.embed_d, config.vocab_size, bias=False)

        self.apply(self._init_weights)
        for pn, p in self.named_parameters():
            if pn.endswith("mlp.2.weight"):
                torch.nn.init.normal_(p, mean=0.0, std=0.02 / sqrt(2 * config.n_layers))

    def _init_weights(self, module: nn.Module) -> None:
        """
        Initialise model weights with GPT2 defaults.

        TODO: Test what difference this makes as cribbed from Karpathy.

        Params:
            module (nn.Module): Module to initialise.
        Returns:
            None
        """

        if isinstance(module, nn.Linear):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias)
        if isinstance(module, nn.MultiheadAttention):
            torch.nn.init.normal_(module.in_proj_weight, mean=0.0, std=0.02)
            torch.nn.init.zeros_(module.in_proj_bias)
            torch.nn.init.normal_(module.out_proj.weight, mean=0.0, std=0.02)
            torch.nn.init.zeros_(module.out_proj.bias)
        elif isinstance(module, nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
        elif isinstance(module, nn.LayerNorm):
            torch.nn.init.zeros_(module.bias)
            torch.nn.init.ones_(module.weight)

    def configure_optimizers(self) -> torch.optim.AdamW:
        """
        Configure AdamW optimizer with some modules experiencing weight decay.

        Returns:
            torch.optim.AdamW: Configured optimizer.
        """

        decay = set()
        for mn, m in self.named_modules():
            if isinstance(m, (torch.nn.Linear, torch.nn.MultiheadAttention)):
                for pn, _ in m.named_parameters():
                    if pn.endswith("weight"):
                        decay.add(f"{mn}.{pn}" if mn else pn)

        param_dict = dict(self.named_parameters())
        all_params = set(param_dict.keys())
        no_decay = all_params - decay
        assert len(decay & no_decay) == 0
        assert len(param_dict.keys() - (decay | no_decay)) == 0

        optim_groups = [
            {
                "params": [param_dict[pn] for pn in sorted(list(decay))],
                "weight_decay": self.config.adam_weight_decay,
            },
            {
                "params": [param_dict[pn] for pn in sorted(list(no_decay))],
                "weight_decay": 0.0,
            },
        ]
        return torch.optim.AdamW(optim_groups, lr=self.config.adam_lr, betas=self.config.adam_betas)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Model forward pass."""
        # pylint: disable=arguments-differ

        _, t = x.size()
        pos = torch.arange(0, t, dtype=torch.long, device=x.device).unsqueeze(0)
        pos_emb = self.transformer.pos_emb(pos)
        tok_emb = self.transformer.tok_emb(x)
        x = self.transformer.dropout(tok_emb + pos_emb)
        for tblock in self.transformer.blocks:
            x = tblock(x)
        x = self.transformer.ln(x)
        return self.lm_head(x)

    def training_step(self, batch: torch.Tensor, batch_idx: torch.Tensor) -> torch.Tensor:
        """
        Lightning training step configuration.

        Params:
            batch (torch.Tensor): Batch of training data.
            batch_idx (torch.Tensor): Index of the batch (UNUSED).
        """
        # pylint: disable=arguments-differ

        _ = batch_idx
        y = batch[:, 1:]
        y_hat = torch.zeros(y.shape[0], 0, self.config.vocab_size, device=y.device)

        for i in range(1, batch.size(1)):
            start = i - self.config.window_size if i >= self.config.window_size else 0
            x = batch[:, start:i]
            x = self.forward(x)
            x = x[:, -1:, :]
            y_hat = torch.cat((y_hat, x), dim=1)

        loss = F.cross_entropy(y_hat.permute(0, 2, 1), y)
        perp = perplexity(y_hat, y)
        self.log("loss", loss, prog_bar=True)
        self.log("perplexity", perp, prog_bar=True)

        return loss

    @torch.no_grad()
    def generate(
        self,
        x: torch.Tensor,
        max_tokens: int,
        t: float = 1.0,
        sample: bool = False,
        top_k: int | None = None,
    ) -> torch.Tensor:
        """
        Generate from a given tokenised input.

        Params:
            x (torch.Tensor): Tokenised input sequence.
            max_tokens (int): Maximum number of tokens to generate.
            t (float): Temperature parameter.
            sample (bool): Sample from the distribution or take the largest probability.
            top_k (int | None): Top-k sampling parameter.
        """
        # pylint: disable=too-many-arguments

        for _ in range(max_tokens):
            if x.size(1) <= self.config.window_size:
                x_trunc = x
            else:
                x_trunc = x[:, -self.config.window_size :]

            logits = self.forward(x_trunc)
            logits = logits[:, -1, :] / t

            if top_k is not None:
                v, _ = torch.topk(logits, top_k)
                logits[logits < v[:, [-1]]] = -float("Inf")

            probs = F.softmax(logits, dim=-1)

            if sample:
                next_ = torch.multinomial(probs, num_samples=1)
            else:
                _, next_ = torch.topk(probs, k=1, dim=-1)

            x = torch.cat((x, next_), dim=-1)
        return x

    def predict_step(self, batch: torch.Tensor, batch_idx: torch.Tensor) -> torch.Tensor:
        """
        Lightning predict step configuration.

        Params:
            batch (torch.Tensor): Batch of training data.
            batch_idx (torch.Tensor): Index of the batch (UNUSED).
        """
        _ = batch_idx

        return self.generate(
            batch,
            max_tokens=self.predict_config.max_tokens,
            t=self.predict_config.t,
            sample=self.predict_config.sample,
            top_k=self.predict_config.top_k,
        )


if __name__ == "__main__":
    pass



"""
Character tokenizers and Lightning DataModule for handling biological datasets.
"""

import gzip
from pathlib import Path

import lightning as L
import numpy as np
import requests
import torch
from Bio import SeqIO
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm


class CharTokenizer:
    """
    Simple single character tokenizer superclass.

    Params:
        vocab: List of characters to be used as the vocabulary.
        pad_char: Character to use for padding.
    Returns:
        None
    """

    def __init__(
        self,
        vocab: list[str],
        pad_char: str = "X",
    ) -> None:
        # pylint: disable=unnecessary-comprehension

        assert pad_char in vocab

        self.vocab = vocab
        self.pad_char = pad_char

        self.encoder = {v: i for i, v in enumerate(vocab)}
        self.decoder = {i: v for i, v in enumerate(vocab)}

    def __call__(self, data: str, pad_to: int | None = None) -> list[int]:
        return self.encode(data, pad_to)

    def __len__(self) -> int:
        return len(self.vocab)

    def encode(self, data: str, pad_to: int | None = None) -> list[int]:
        """
        Call the encoder, padding and cropping to the desired length.

        Params:
            data (str): Sequence to encode.
            pad_to (int): Length to pad or crop to.
        Returns:
            list[int]: Encoded sequence.
        """
        if pad_to is not None:
            data = data[:pad_to]
            data += self.pad_char * (pad_to - len(data))
        return [self.encoder[c] for c in data]

    def decode(self, data: torch.Tensor) -> str:
        """
        Decode an encoded sequence.

        TODO: Correct type hint.

        Params:
            data (torch.Tensor): Encoded sequence.
        Returns:
            str: Decoded sequence.
        """
        return "".join([self.decoder[int(i)] for i in data])


class NTTokenizer(CharTokenizer):
    """
    DNA nucleotide character tokenizer.

    Uses IUPAC nucleotide codes and `N` for padding. (Vocab size = 5)
    """

    def __init__(self) -> None:
        super().__init__(list("ATCGN"), "N")


class AATokenizer(CharTokenizer):
    """
    Protein amino acid character tokenizer.

    Uses standard AAs, `ZBU` non-canonical AAs and `X` for padding.
    (Vocab size = 26)
    """

    def __init__(self) -> None:
        super().__init__(list("ACDEFGHIKLMNPQRSTVWYZBUX"), "X")


class FastaDataset(Dataset):
    """
    Pytorch Dataset for loading, encoding and serving sequences from fasta files.

    Params:
        directory (Path): Path to a directory containing .fa or .fasta files (can be gzipped).
        tokenizer (CharTokenizer): Tokenizer to use for encoding sequences.
        max_len (int): Maximum sequence length to crop and pad sequences to.
        limit (int): Maximum number of sequences to load.
        bin_name (Path): Path to a binary file to where the encoded data is stored.
    Returns:
        None

    Note: If `bin_name` already exists, it will be overwritten.
    """

    def __init__(
        self,
        directory: Path,
        tokenizer: CharTokenizer,
        max_len: int,
        limit: int,
        bin_name: Path = Path("train.bin"),
    ) -> None:
        # pylint: disable=too-many-arguments,consider-using-with, unspecified-encoding

        print("Initializing FastaDataset...")
        self.files = (
            list(directory.glob("*.fasta"))
            + list(directory.glob("*.fa"))
            + list(directory.glob("*.fasta.gz"))
            + list(directory.glob("*.fa.gz"))
        )
        self.bin_name = bin_name
        self.tokenizer = tokenizer
        self.max_len = max_len

        assert np.iinfo(np.uint8).max >= len(tokenizer)  # Ensure vocab fits into 256 tokens

        # Init memmap to hold the maximum number of sequences of length max_len:
        self.data = np.memmap(self.bin_name, dtype=np.uint8, mode="w+", shape=(limit, max_len))

        # Loop through files and sequences, tokenise and store in the memmap:
        n_seqs = 0
        for file in self.files:
            print(f"Loading: {file}...")

            if file.suffix == ".gz":
                handle = gzip.open(file, "rt")
            else:
                handle = open(file, "r")

            for record in SeqIO.parse(handle, "fasta"):
                seq = str(record.seq).upper()[:max_len]
                self.data[n_seqs] = tokenizer(seq, max_len)
                n_seqs += 1
                if n_seqs >= limit:
                    break

            handle.close()

        print(f"Finished: loaded {n_seqs} sequences ({max_len*n_seqs} tokens)")
        self._len = n_seqs

        # Resize the memmap to the actual number of sequences:
        self.data = np.memmap(self.bin_name, dtype=np.uint8, mode="r+", shape=(n_seqs, max_len))

    def __len__(self) -> int:
        return self._len

    def __getitem__(self, idx: int) -> torch.Tensor:
        data = self.data[idx]
        return torch.LongTensor(data)

    def cleanup(self) -> None:
        """
        Flush and remove the dataset mmap binary.
        """
        self.data.flush()
        self.bin_name.unlink()


class InferenceDataset(Dataset):
    """Simple dataset to serve the model the tokenised prompt during inference."""

    def __init__(self, prompt: str, tokenizer: CharTokenizer, n_samples) -> None:
        self.prompt = prompt
        self.tokenizer = tokenizer

        self.x = [self.tokenizer(self.prompt)] * n_samples

    def __len__(self) -> int:
        return len(self.x)

    def __getitem__(self, idx: int) -> torch.Tensor:
        return torch.LongTensor(self.x[idx])


class BioDataModule(L.LightningDataModule):
    """
    Pytorch Lightning DataModule for loading, encoding and serving sequences from fasta files.

    Params:
        mode (str): Either "nt" or "aa" to indicate the type of sequences to loadu
        directory (Path): Path to a directory containing .fa or .fasta files (can be gzipped).
        batch_size (int): Batch size to use for training.
        max_seq_length (int): Maximum sequence length to crop and pad sequences to.
        n_seq_limit (int): Maximum number of sequences to load.
        loader_num_workers (int): Number of workers to use for the torch DataLoader.
        downloads (list[tuple[str, str]]): List of tuples containing URLs and filenames to download.
    Returns:
        None
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        mode: str,
        directory: Path,
        batch_size: int,
        max_seq_length: int,
        n_seq_limit: int,
        loader_num_workers: int,
        downloads: list[tuple[str, str]] | None,
        prompt: str | None = None,
        n_samples: int | None = None,
    ) -> None:

        # pylint: disable=too-many-arguments
        super().__init__()

        assert mode in ["nt", "aa"]

        self.mode = mode
        self.directory = directory
        self.batch_size = batch_size
        self.max_seq_length = max_seq_length
        self.n_seq_limit = n_seq_limit
        self.loader_num_workers = loader_num_workers
        self.downloads = downloads

        # Predict mode settings:
        self.prompt = prompt
        self.n_samples = n_samples

        if self.downloads is not None:
            for url, filename in self.downloads:
                self.download_dataset(url, filename)

        self.tokenizer = NTTokenizer() if self.mode == "nt" else AATokenizer()
        self.train_dataset = FastaDataset(
            self.directory, self.tokenizer, self.max_seq_length, self.n_seq_limit
        )

    def download_dataset(self, url: str, filename: str, chunk_size: int = 1024 * 10) -> None:
        """
        Download a file from a URL to a local file.

        Params:
            url (str): URL to download from.
            filename (str): Filename to save the file as.
            chunk_size (int): Chunk size to use for downloading.
        Returns:
                None
        """
        assert self.downloads is not None
        download_path = self.directory / filename

        if download_path.exists():
            print(f"Downloaded dataset already exists: {download_path}")
            return

        print(f"Downloading dataset from: {url}")
        self.directory.mkdir(exist_ok=True)
        response = requests.get(url, stream=True, timeout=None)

        size_mb = int(response.headers.get("content-length", 0)) / (1024**2)
        progress_bar = tqdm(total=int(size_mb), unit="MB")
        with open(download_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:
                    f.write(chunk)
                    progress_bar.update(chunk_size / (1024**2))

    def train_dataloader(self) -> DataLoader:
        """Return the DataLoader for training."""
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            num_workers=self.loader_num_workers,
        )

    def teardown(self, stage: str | None = None) -> None:
        """Call the cleanup method of the train dataset."""
        _ = stage
        self.train_dataset.cleanup()

    def predict_dataloader(self) -> DataLoader:
        """Return the DataLoader for inference."""
        assert self.prompt is not None

        dataset = InferenceDataset(self.prompt, self.tokenizer, self.n_samples)
        return DataLoader(dataset, batch_size=1, num_workers=self.loader_num_workers)


if __name__ == "__main__":
    pass
